
body {
    font-family: "Droid Sans",Helvetica,Arial,sans-serif;
    background-color: #ffffff;
    color: #202060;
}

.tabs, .tabs2, .tabs3, .navpath ul, .tablist li {
    background-image: none;
}

.tabs, .tabs2, .tabs3 {
    border-top: 1px solid #202060;
}

div.contents {
    margin: 0px;
    padding-top: 10px;
    padding-left: 12px;
    padding-right: 8px;
}

